#include <stdio.h>
#include <errno.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <poll.h>

#include "../src/include/netfilter_platform.h"

/* poll timeout 500ms */
#define POLL_RECV_TIMEOUT (500)


void* receive_dummy_packet(void* arg)
{
    int ret = -1;
    int sock = 0;
    int enable = 1;
    int new_socket = 0;
    char buffer[10] = {0};
    struct sockaddr_in addr;
    int addr_len = sizeof(addr);

    sock = socket(AF_INET, SOCK_STREAM, 0);
    CHECK_ERROR(sock < 0, "Failed to create socket: %s", strerror(errno));

    ret = setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(int));
    CHECK_ERROR(ret < 0, "Failed to setsockopt SO_REUSEADDR: %s", strerror(errno));

    addr.sin_family = AF_INET;
    addr.sin_port = htons(12345);
    addr.sin_addr.s_addr = INADDR_ANY;

    /* Binding to an address */
    ret = bind(sock, (struct sockaddr *)&addr, sizeof(addr));
    CHECK_ERROR(ret < 0, "Bind failed: %s", strerror(errno));

    /* Listening with 3 connection limit */
    ret = listen(sock, 3);
    CHECK_ERROR(ret < 0, "Failed to listen: %s", strerror(errno));

    /* accepting - waiting for client to connect */
    new_socket = accept(sock, (struct sockaddr *)&addr, (socklen_t*)&addr_len);
    CHECK_ERROR(new_socket < 0, "Failed to accept: %s", strerror(errno));

    /* reading messages from the client */
    ret = read(new_socket , buffer, 10);
    CHECK_ERROR(ret == -1, "Failed to read from socket: %s", strerror(errno));

    printf("Message received: %s\n", buffer);

exit:
	if (sock >= 0)
	{
		close(sock);
	}
    if (new_socket >= 0)
	{
		close(new_socket);
	}
	return NULL;
}

void* create_and_send_dummy_packet(void* arg)
{
	int ret = -1;
    int sock = 0;
    struct sockaddr_in addr;

    // Create the socket.
    sock = socket(AF_INET, SOCK_STREAM, 0);
    CHECK_ERROR(sock < 0, "Failed to create socket: %s", strerror(errno));

    // Bind a name to the socket.
    addr.sin_family = AF_INET;
    addr.sin_port = htons(12345);
    // Convert IPv4 address from text to binary form
    ret = inet_pton(AF_INET, "127.0.0.1", &addr.sin_addr);
    CHECK_ERROR(ret != 1, "Invalid address/ Address not supported");

    // let the start reading on the socket
	sleep(1);
    ret = connect(sock, (struct sockaddr *)&addr, sizeof(addr));
    CHECK_ERROR(ret < 0, "Connection failed %s", strerror(errno));

    ret = send(sock, "hello", 5, 0);
    CHECK_ERROR(ret < 0, "failed to send message");

	ret = 0;
    printf("Sent %d bytes successfully\n", ret);

exit:
	if (sock >= 0)
	{
		close(sock);
	}
	return NULL;
}

void packet_vetting(int fd, int response)
{
	ssize_t ret;
	struct conn_identity_data conn_data;
	packet_verdict pkt_verdict = {0, NF_REPEAT};
	fprintf(stderr, "FD of parent process: %d.\n", fd);
    struct pollfd poll_fd = {0};

    /* setup poll */
    poll_fd.events = POLLIN;
    poll_fd.fd = fd;
    poll_fd.revents = 0;

    while(1)
    {
        ret = poll(&poll_fd, 1, POLL_RECV_TIMEOUT);
        if (ret < 0 && errno != EINTR)  // interrupted, continue
        {
            ERROR("Poll interrupted: %s", strerror(errno));
            continue;
        }

        if (ret == 0) // timeout
        {
            continue;
        }
        else if (poll_fd.revents & POLLIN) // we have data to read
        {
            memset(&conn_data, 0, sizeof(struct conn_identity_data));

            ret = read(fd, &conn_data, sizeof(struct conn_identity_data));
            CHECK_ERROR(ret <= 0, "Failed to read packet from the server: %s", strerror(errno));
            printf("Received packet: %d\t protocol: %d\t event_type: %d\n", conn_data.event_id, conn_data.protocol, conn_data.event_type);

            /* prepare verdict */
            pkt_verdict.packetid = conn_data.event_id;
            if (response == NF_STOP)
            {
                printf("Timeout to send back verdict\n");
                sleep(2); // Timeout reached for sending verdict
                pkt_verdict.verdict = NF_REPEAT;
            }
            else
            {
                pkt_verdict.verdict = response;
            }

            ret = write(fd, &pkt_verdict, sizeof(packet_verdict));
            CHECK_ERROR(ret <= 0, "Failed to send verdict back to the server: %s", strerror(errno));
            printf("Sent verdict back for PacketID = %d Verdict = %d \n", pkt_verdict.packetid, pkt_verdict.verdict);
            sleep(1);
            break;
        }
    }

exit:
	close(fd);
}

void usage(char* arg)
{
    printf("\t\"\"Usage: %s <verdict>\n"
            "\t<verdict> can be one of the following values\n"
            "\tallow : To allow the packet sent from client to server\n"
            "\tdrop: To drop the packet sent from client to server\n"
            "\ttimeout: Failed to pass back verdict in time, packet allowed\"\"\n\n\n", arg);
    return;
}

int main(int argc, char ** argv) 
{
	int ret = -1;
    int sock = 0;
    struct sockaddr_un name = {0};
    const char* sock_name = "/var/run/mde_netfilter.sock";
    pthread_t network_event_gen_thread;
    pthread_t network_event_listener_thread;
    int response = NF_REPEAT;

    if (argc == 2)
    {
        if (strcmp(argv[1], "allow") == 0)
        {
            response = NF_REPEAT;
        }
        else if (strcmp(argv[1], "drop") == 0)
        {
            response = NF_DROP;
        }
        else if (strcmp(argv[1], "timeout") == 0)
        {
            response = NF_STOP;
        }
        else
        {
            printf("Incorrect argument\n");
            usage(argv[0]);
            exit(EXIT_FAILURE);
        }
    }
    else
    {
        usage(argv[0]);
        exit(EXIT_FAILURE);
    }

    /* Create the socket. */
    sock = socket(AF_UNIX, SOCK_STREAM, 0);
    if (sock < 0) {
        perror("socket");
        exit(EXIT_FAILURE);
    }

    /* Bind a name to the socket. */
    name.sun_family = AF_UNIX;
    strncpy(name.sun_path, sock_name, sizeof (name.sun_path));
    name.sun_path[sizeof (name.sun_path) - 1] = '\0';

    /* connect to server */
    if (connect(sock, (struct sockaddr *)&name, sizeof(name)) == -1) {
      perror("Failed to connect socket. Error:");
      exit(EXIT_FAILURE);
    }

    /* create receiver */
    ret = pthread_create(&network_event_listener_thread, NULL, receive_dummy_packet, NULL);
    CHECK_ERROR(ret < 0, "Failed to create receiver. Abort");

    /* create and send packet */
    ret = pthread_create(&network_event_gen_thread, NULL, create_and_send_dummy_packet, NULL);
    CHECK_ERROR(ret < 0, "Failed to create and send packet. Abort");

    /* receive packet from server and provide verdict */
    packet_vetting(sock, response);

exit:
    if (sock >= 0)
    {
        close(sock);
    }
	return EXIT_SUCCESS;
}